
// =================================================================================================
// This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. It
// is auto-generated by the 'scripts/database/database.py' Python script.
//
// This file populates the database with best-found tuning parameters for the 'Copy' kernels.
//
// =================================================================================================

#include "database/kernels/copy/copy.hpp"

#include "database/kernels/copy/copy_16.hpp"
#include "database/kernels/copy/copy_32.hpp"
#include "database/kernels/copy/copy_3232.hpp"
#include "database/kernels/copy/copy_64.hpp"
#include "database/kernels/copy/copy_6464.hpp"
