
// =================================================================================================
// This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. It
// is auto-generated by the 'scripts/database/database.py' Python script.
//
// This file populates the database with best-found tuning parameters for the 'Xaxpy' kernels.
//
// =================================================================================================

#include "database/database_structure.hpp"

namespace clblast {
namespace database {

extern const DatabaseEntry XaxpyHalf;
extern const DatabaseEntry XaxpySingle;
extern const DatabaseEntry XaxpyComplexSingle;
extern const DatabaseEntry XaxpyDouble;
extern const DatabaseEntry XaxpyComplexDouble;

} // namespace database
} // namespace clblast
