
// =================================================================================================
// This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. It
// is auto-generated by the 'scripts/database/database.py' Python script.
//
// This file populates the database with best-found tuning parameters for the 'Xconvgemm' kernels.
//
// =================================================================================================

#include "database/kernels/xconvgemm/xconvgemm.hpp"

#include "database/kernels/xconvgemm/xconvgemm_16.hpp"
#include "database/kernels/xconvgemm/xconvgemm_32.hpp"
#include "database/kernels/xconvgemm/xconvgemm_3232.hpp"
#include "database/kernels/xconvgemm/xconvgemm_64.hpp"
#include "database/kernels/xconvgemm/xconvgemm_6464.hpp"
