
// =================================================================================================
// This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. It
// is auto-generated by the 'scripts/database/database.py' Python script.
//
// This file populates the database with best-found tuning parameters for the 'Trsv_Routine16' kernels.
//
// =================================================================================================
#include "database/kernels/trsv_routine/trsv_routine.hpp"
namespace clblast {
namespace database {

const DatabaseEntry TrsvRoutineHalf = {
  "TrsvRoutine", Precision::kHalf, {"TRSV_BLOCK_SIZE"}, {
    { // Default
      kDeviceTypeAll, "default", {
        { "default", {
          { kDeviceNameDefault                                        , Params{ 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } },
        } },
      }
    },
  }
};

} // namespace database
} // namespace clblast
